import pytest
from opr import Primer, MeltingTemperature, OPRBaseError

TEST_CASE_NAME = "Errors tests"


def test_addition():
    oprimer = Primer("ATCGATCGATCGATCGAT")
    with pytest.raises(OPRBaseError, match=r"You can only add two Primer objects."):
        oprimer + 2


def test_multiply():
    oprimer = Primer("ATCGATCGATCGATCGAT")
    with pytest.raises(OPRBaseError, match=r"The primer sequence can only be multiplied by an integer."):
        oprimer * "2"


def test_validate_primer_1():
    with pytest.raises(OPRBaseError, match=r"Primer sequence should be a string variable."):
        oprimer = Primer(222)


def test_validate_primer_2():
    with pytest.raises(OPRBaseError, match=r"Primer sequence should only contain the nucleotide bases A, T, C, and G."):
        oprimer = Primer("ATCGATCGATCGATCGAF")


def test_melting_temperature_1():
    oprimer = Primer("ATCGATCGATCGATCGAT")
    with pytest.raises(NotImplementedError, match=r"This method for calculating melting temperature has not been implemented."):
        oprimer.melting_temperature(method=MeltingTemperature.NEAREST_NEIGHBOR)


def test_melting_temperature_2():
    oprimer = Primer("ATCGATCGATCGATCGAT")
    with pytest.raises(NotImplementedError, match=r"This method for calculating melting temperature has not been implemented."):
        oprimer.melting_temperature(method="Invalid Method")


def test_is_computed():
    oprimer = Primer("ATCGATCGATCGATCGAT")
    with pytest.raises(OPRBaseError, match=r"^This attribute either doesn't exist or cannot be computed/cached."):
        oprimer.is_computed("Invalid Attribute")



from opr import Primer

TEST_CASE_NAME = "Operations tests"


def test_reverse_1():  # Reference: https://www.qiagen.com/us/applications/enzymes/tools-and-calculators/reverse-complement-converter
    oprimer = Primer("ATCGATCGATCGATCGAT")
    oprimer_reversed = oprimer.reverse(inplace=False)
    assert oprimer_reversed.sequence == "TAGCTAGCTAGCTAGCTA"


def test_reverse_2():  # Reference: https://www.qiagen.com/us/applications/enzymes/tools-and-calculators/reverse-complement-converter
    oprimer = Primer("ATCGATCGATCGATCGAT")
    oprimer_reversed = oprimer.reverse()
    assert oprimer_reversed.sequence == "TAGCTAGCTAGCTAGCTA"


def test_reverse_3():  # Reference: https://www.qiagen.com/us/applications/enzymes/tools-and-calculators/reverse-complement-converter
    oprimer = Primer("ATCGATCGATCGATCGAT")
    oprimer.reverse(inplace=True)
    assert oprimer.sequence == "TAGCTAGCTAGCTAGCTA"


def test_complement_1():  # Reference: https://www.qiagen.com/us/applications/enzymes/tools-and-calculators/reverse-complement-converter
    oprimer = Primer("ATCGGCTAAATCGGCTAA")
    oprimer_complemented = oprimer.complement(inplace=False)
    assert oprimer_complemented.sequence == "TAGCCGATTTAGCCGATT"


def test_complement_2():  # Reference: https://www.qiagen.com/us/applications/enzymes/tools-and-calculators/reverse-complement-converter
    oprimer = Primer("ATCGGCTAAATCGGCTAA")
    oprimer_complemented = oprimer.complement()
    assert oprimer_complemented.sequence == "TAGCCGATTTAGCCGATT"


def test_complement_3():  # Reference: https://www.qiagen.com/us/applications/enzymes/tools-and-calculators/reverse-complement-converter
    oprimer = Primer("ATCGGCTAAATCGGCTAA")
    oprimer.complement(inplace=True)
    assert oprimer.sequence == "TAGCCGATTTAGCCGATT"


def test_to_rna_1():  # Reference: https://biomodel.uah.es/en/lab/cybertory/analysis/trans.htm
    oprimer = Primer("ATCGGCTAAATCGGCTAA")
    oprimer_rna = oprimer.to_rna()
    assert oprimer_rna == "AUCGGCUAAAUCGGCUAA"


def test_to_rna_2():  # Reference: https://biomodel.uah.es/en/lab/cybertory/analysis/trans.htm
    oprimer = Primer("ATCGATCGATCG")
    oprimer_rna = oprimer.to_rna()
    assert oprimer_rna == "AUCGAUCGAUCG"


def test_length():
    oprimer = Primer("ATCGGCTAAATCGGCTAA")
    assert len(oprimer) == 18


def test_addition():
    oprimer_1 = Primer("ATCG")
    oprimer_2 = Primer("GATC")
    oprimer_concat = oprimer_1 + oprimer_2
    assert oprimer_concat.sequence == "ATCGGATC"


def test_multiply():
    oprimer_1 = Primer("ATCG")
    oprimer_concat = oprimer_1 * 4
    assert oprimer_concat.sequence == "ATCGATCGATCGATCG"


def test_equality1():
    oprimer_1 = Primer("ATCG")
    oprimer_2 = Primer("ATCG")
    assert oprimer_1 == oprimer_2


def test_equality2():
    oprimer_1 = Primer("ATCG")
    oprimer_2 = Primer("ATCGC")
    assert oprimer_1 != oprimer_2


def test_equality3():
    oprimer = Primer("ATCG")
    assert oprimer != 2


def test_str():
    oprimer_1 = Primer("ATCG")
    oprimer_2 = Primer("ATCGC")
    oprimer_concat = oprimer_1 + oprimer_2
    assert str(oprimer_1) + str(oprimer_2) == oprimer_concat.sequence


def test_iter():
    oprimer_1 = Primer("ATCG")
    sequence = ""
    for base in oprimer_1:
        sequence += base
    assert oprimer_1.sequence == sequence


def test_contains1():
    oprimer_1 = Primer("ATCG")
    oprimer_2 = Primer("ATCGAT")
    assert oprimer_1 in oprimer_2


def test_contains2():
    sequence = "ATCG"
    oprimer = Primer("ATCGAT")
    assert sequence in oprimer


def test_contains3():
    sequence = "ATCG"
    oprimer = Primer("TCGAT")
    assert sequence not in oprimer


def test_contains4():
    oprimer = Primer("TCGAT")
    assert 2 not in oprimer



import itertools
from opr.params import VALID_BASES
from opr import Primer, MeltingTemperature

TEST_CASE_NAME = "Cache tests"


def test_mwc():
    oprimer = Primer("ATCGATCGATCGATCGAT")
    assert not oprimer.is_computed("molecular_weight")
    molecular_weight_first = oprimer.molecular_weight
    assert oprimer.is_computed("molecular_weight")

    molecular_weight_second = oprimer.molecular_weight
    assert round(molecular_weight_first, 1) == round(molecular_weight_second, 1)


def test_gc_content():
    oprimer = Primer("ATTCG")
    assert not oprimer.is_computed("gc_content")
    gc_content_first = oprimer.gc_content
    assert oprimer.is_computed("gc_content")

    gc_content_second = oprimer.gc_content
    assert gc_content_first == gc_content_second


def test_gc_clamp():
    oprimer = Primer("ATCGATCGATCGATCGGTCG")
    assert not oprimer.is_computed("gc_clamp")
    gc_clamp_first = oprimer.gc_clamp
    assert oprimer.is_computed("gc_clamp")

    gc_clamp_second = oprimer.gc_clamp
    assert gc_clamp_first == gc_clamp_second


def test_melt_temp():
    oprimer = Primer("ATCGATCGATCGATCGATCG")
    assert not oprimer.is_computed("melting_temperature")[MeltingTemperature.BASIC]
    basic_melt_temp_first = oprimer.melting_temperature(MeltingTemperature.BASIC)
    assert oprimer.is_computed("melting_temperature")[MeltingTemperature.BASIC]

    basic_melt_temp_second = oprimer.melting_temperature(MeltingTemperature.BASIC)
    assert round(basic_melt_temp_first, 1) == round(basic_melt_temp_second, 1)


def test_single_runs():
    oprimer = Primer("AAAAATTCGGGGATCCCCG")
    assert not oprimer.is_computed("single_runs")
    runs_first = oprimer.single_runs
    assert oprimer.is_computed("single_runs")

    runs_second = oprimer.single_runs
    assert all(runs_first[pair] == runs_second[pair] for pair in runs_first)
    assert all(runs_second[pair] == runs_first[pair] for pair in runs_second)


def test_double_runs():
    p1 = Primer("ATATCGAACACACACACA")
    assert not p1.is_computed("double_runs")
    double_runs_first = p1.double_runs
    assert p1.is_computed("double_runs")

    double_runs_second = p1.double_runs
    assert len(double_runs_second) == len(double_runs_first)
    assert all(double_runs_first[pair] == double_runs_second[pair] for pair in double_runs_first)
    assert all(double_runs_second[pair] == double_runs_first[pair] for pair in double_runs_second)


def test_e260():
    oprimer = Primer("ATCGATCGATCGATCGATCG")
    assert not oprimer.is_computed("E260")
    e260_first = oprimer.E260
    assert oprimer.is_computed("E260")

    e260_second = oprimer.E260
    assert round(e260_first, 1) == round(e260_second, 1)



# -*- coding: utf-8 -*-
"""OPR primer."""
import re
import itertools
from enum import Enum
from warnings import warn
from .errors import OPRBaseError
from .params import DEFAULT_PRIMER_NAME, VALID_BASES
from .params import PRIMER_SEQUENCE_TYPE_ERROR, PRIMER_SEQUENCE_LENGTH_WARNING, PRIMER_SEQUENCE_VALID_BASES_ERROR, PRIMER_SEQUENCE_VALID_GC_CONTENT_RANGE_WARNING
from .params import PRIMER_LOWER_LENGTH, PRIMER_HIGHEST_LENGTH, PRIMER_LOWEST_GC_RANGE, PRIMER_HIGHEST_GC_RANGE
from .params import DNA_COMPLEMENT_MAP
from .params import PRIMER_ADDITION_ERROR, PRIMER_MULTIPLICATION_ERROR
from .params import PRIMER_MELTING_TEMPERATURE_NOT_IMPLEMENTED_ERROR
from .params import PRIMER_ATTRIBUTE_NOT_COMPUTABLE_ERROR
from .functions import molecular_weight_calc, basic_melting_temperature_calc, salt_adjusted_melting_temperature_calc, gc_clamp_calc
from .functions import e260_ssnn_calc


class MeltingTemperature(Enum):
    """Mode used to calculate the Melting Temperature of the Primer accordingly."""

    BASIC = 1
    SALT_ADJUSTED = 2
    NEAREST_NEIGHBOR = 3


class Primer:
    """
    The Primer class facilitates working with the primer sequence.

    >>> oprimer = Primer("ATCGATCGATCGATCGAT")
    >>> oprimer.molecular_weight
    """

    def __init__(self, sequence, name=DEFAULT_PRIMER_NAME, salt=50):
        """
        Initialize the Primer instance.

        :param sequence: primer nucleotides sequence
        :type sequence: str
        :param name: primer name
        :type name: str
        :param salt: Sodium ion concentration in moles (unit mM)
        :type salt: float
        :return: an instance of the Primer class
        """
        self._sequence = Primer.validate_primer(sequence)
        self._name = name
        self._molecular_weight = None
        self._gc_content = None
        self._gc_clamp = None
        self._single_runs = None
        self._double_runs = None
        self._E260 = None
        self._salt_level = salt
        self._melting_temperature = {
            MeltingTemperature.BASIC: None,
            MeltingTemperature.SALT_ADJUSTED: None,
            MeltingTemperature.NEAREST_NEIGHBOR: None,
        }

        # Track computed attributes
        self._computed = {
            "molecular_weight": False,
            "gc_content": False,
            "gc_clamp": False,
            "single_runs": False,
            "double_runs": False,
            "E260": False,
            "melting_temperature": {
                MeltingTemperature.BASIC: False,
                MeltingTemperature.SALT_ADJUSTED: False,
                MeltingTemperature.NEAREST_NEIGHBOR: False,
            },
        }

    def is_computed(self, attr):
        """
        Check whether the given attribute has been computed.

        :param attr: The attribute to check.
        :type attr: str
        :return: True if the attribute is previously computed, False otherwise.
        """
        if attr not in self._computed:
            raise OPRBaseError(PRIMER_ATTRIBUTE_NOT_COMPUTABLE_ERROR)
        return self._computed.get(attr, False)

    def __len__(self):
        """
        Return the length of the Primer sequence.

        :return: length of the Primer sequence
        """
        return len(self._sequence)

    def __eq__(self, other_primer):
        """
        Check primers equality.

        :param other_primer: another Primer
        :type other_primer: Primer
        :return: result as bool
        """
        if isinstance(other_primer, Primer):
            return self._sequence == other_primer._sequence
        return False

    def __add__(self, other_primer):
        """
        Concatenate the sequences of the current Primer with another one.

        :param other_primer: another Primer to concat its sequence to the current Primer
        :type other_primer: Primer
        :return: new Primer with concatenated sequence
        """
        if isinstance(other_primer, Primer):
            return Primer(self._sequence + other_primer._sequence)
        raise OPRBaseError(PRIMER_ADDITION_ERROR)

    def __mul__(self, number):
        """
        Multiply the Primer sequence `number` times.

        :param number: times to concat the Primer sequence to itself
        :type number: int
        :return: new Primer with multiplied sequence
        """
        if isinstance(number, int):
            return Primer(self._sequence * number)
        raise OPRBaseError(PRIMER_MULTIPLICATION_ERROR)

    def __contains__(self, sequence):
        """
        Check if the Primer contains the given sequence.

        :param sequence: sequence
        :type sequence: str/Primer
        :return: bool
        """
        if isinstance(sequence, str):
            return sequence in self._sequence
        elif isinstance(sequence, Primer):
            return sequence._sequence in self._sequence
        return False

    def __str__(self):
        """
        Primer object string representation method.

        :return: primer sequence as str
        """
        return self._sequence

    def __iter__(self):
        """
        Iterate through Primer.

        :return: base as Generator[str]
        """
        yield from self.sequence

    def reverse(self, inplace=False):
        """
        Reverse sequence.

        :param inplace: inplace flag
        :type inplace: bool
        :return: new Primer object or None
        """
        new_sequence = self._sequence[::-1]
        if inplace:
            self._sequence = new_sequence
        else:
            return Primer(sequence=new_sequence)

    def complement(self, inplace=False):
        """
        Complement sequence.

        :param inplace: inplace flag
        :type inplace: bool
        :return: new Primer object or None
        """
        new_sequence = ""
        for item in self._sequence:
            new_sequence += DNA_COMPLEMENT_MAP[item]
        if inplace:
            self._sequence = new_sequence
        else:
            return Primer(sequence=new_sequence)

    def to_rna(self):
        """
        Convert DNA sequence to RNA.

        :return: str
        """
        return self._sequence.replace('T', 'U')

    @staticmethod
    def validate_primer(sequence):
        """
        Validate the given primer sequence.

        :param sequence: primer nucleotides sequence
        :type sequence: any
        :return: an uppercased primer sequence
        """
        if not isinstance(sequence, str):
            raise OPRBaseError(PRIMER_SEQUENCE_TYPE_ERROR)
        sequence = sequence.upper()

        if len(sequence) < PRIMER_LOWER_LENGTH or len(sequence) > PRIMER_HIGHEST_LENGTH:
            warn(PRIMER_SEQUENCE_LENGTH_WARNING, RuntimeWarning)

        if not all(base in VALID_BASES for base in sequence):
            raise OPRBaseError(PRIMER_SEQUENCE_VALID_BASES_ERROR)
        return sequence

    @property
    def sequence(self):
        """
        Return the primer sequence.

        :return: primer sequence
        """
        return self._sequence

    @property
    def name(self):
        """
        Return the primer name.

        :return: primer name
        """
        return self._name

    @property
    def molecular_weight(self):
        """
        Calculate(if needed) the molecular weight.

        :return: molecular weight
        """
        if not self._computed["molecular_weight"]:
            self._molecular_weight = molecular_weight_calc(self._sequence)
            self._computed["molecular_weight"] = True
        return self._molecular_weight

    @property
    def gc_content(self):
        """
        Calculate gc content.

        :return: gc content
        """
        if not self._computed["gc_content"]:
            gc_count = self._sequence.count('G') + self._sequence.count('C')
            self._gc_content = gc_count / len(self._sequence)
            self._computed["gc_content"] = True
        if self._gc_content < PRIMER_LOWEST_GC_RANGE or self._gc_content > PRIMER_HIGHEST_GC_RANGE:
            warn(PRIMER_SEQUENCE_VALID_GC_CONTENT_RANGE_WARNING, RuntimeWarning)
        return self._gc_content

    @property
    def gc_clamp(self):
        """
        Calculate GC clamp of the primer.

        :return: GC clamp of the primer
        """
        if not self._computed["gc_clamp"]:
            self._gc_clamp = gc_clamp_calc(self._sequence)
            self._computed["gc_clamp"] = True
        return self._gc_clamp

    @property
    def single_runs(self):
        """
        Calculate Single Runs of the primer.

        Run length refers to how many times a single base is repeated consecutively in the primer.

        :return: single runs of the primer
        """
        if not self._computed["single_runs"]:
            self._single_runs = {}
            for base in VALID_BASES:
                self._single_runs[base] = self.repeats(base, consecutive=True)
            self._computed["single_runs"] = True
        return self._single_runs

    @property
    def double_runs(self):
        """
        Calculate Double Runs of the primer.

        It refers to how many times each 2-base pairs occurs consecutively in the primer.

        :return: Dictionary of double runs (2-base pairs) and their counts in the primer
        """
        if not self._computed["double_runs"]:
            pairs = [''.join(pair) for pair in itertools.product(VALID_BASES, repeat=2) if pair[0] != pair[1]]
            counts = {}
            for pair in pairs:
                counts[pair] = self.repeats(pair, consecutive=True)
            self._double_runs = counts
            self._computed["double_runs"] = True
        return self._double_runs

    @property
    def E260(self):
        """
        Calculate the extinction coefficient at 260 nm.

        :return: extinction coefficient at 260 nm
        """
        if not self._computed["E260"]:
            self._E260 = e260_ssnn_calc(self._sequence)
            self._computed["E260"] = True
        return self._E260

    def repeats(self, sequence, consecutive=False):
        """
        Count occurrences of a subsequence in a given sequence.

        :param sequence: The sequence to search within.
        :type sequence: str
        :param consecutive: Whether to count only consecutive repeats.
        :type consecutive: bool
        :return: The count of occurrences.
        """
        if consecutive:
            pattern = f"(?:{re.escape(sequence)})+"
            matches = re.findall(f"({pattern})+", self.sequence)
            result = max((len(match) // len(sequence) for match in matches), default=0)
            return result
        else:
            return self.sequence.count(sequence)

    def melting_temperature(self, method=MeltingTemperature.BASIC):
        """
        Calculate(if needed) the melting temperature.

        :param method: requested calculation mode for melting temperature
        :type method: MeltingTemperature
        :return: approximated melting temperature
        """
        if method not in self._computed["melting_temperature"]:
            raise NotImplementedError(PRIMER_MELTING_TEMPERATURE_NOT_IMPLEMENTED_ERROR)
        if self._computed["melting_temperature"][method]:
            return self._melting_temperature[method]

        if method == MeltingTemperature.BASIC:
            self._melting_temperature[MeltingTemperature.BASIC] = basic_melting_temperature_calc(self._sequence)
        elif method == MeltingTemperature.SALT_ADJUSTED:
            self._melting_temperature[MeltingTemperature.SALT_ADJUSTED] = salt_adjusted_melting_temperature_calc(
                self._sequence, self._salt_level)
        else:
            raise NotImplementedError(PRIMER_MELTING_TEMPERATURE_NOT_IMPLEMENTED_ERROR)
        self._computed["melting_temperature"][method] = True
        return self._melting_temperature[method]



# -*- coding: utf-8 -*-
"""OPR functions."""
import math
from .params import A_WEIGHT, T_WEIGHT, C_WEIGHT, G_WEIGHT, ANHYDROUS_MOLECULAR_WEIGHT_CONSTANT
from .params import BASE_EXTINCTION_COEFFICIENTS, NN53_EXTINCTION_COEFFICIENTS


def molecular_weight_calc(sequence):
    """
    Calculate molecular weight.

    :param sequence: primer nucleotides sequence
    :type sequence: str
    :return: molecular weight as float
    """
    a_count = sequence.count('A')
    t_count = sequence.count('T')
    c_count = sequence.count('C')
    g_count = sequence.count('G')
    return (a_count * A_WEIGHT) + (t_count * T_WEIGHT) + (c_count * C_WEIGHT) + \
        (g_count * G_WEIGHT) - ANHYDROUS_MOLECULAR_WEIGHT_CONSTANT


def basic_melting_temperature_calc(sequence):
    """
    Calculate basic melting temperature.

    :param sequence: primer nucleotides sequence
    :type sequence: str
    :return: melting temperature as float
    """
    a_count = sequence.count('A')
    t_count = sequence.count('T')
    c_count = sequence.count('C')
    g_count = sequence.count('G')
    if len(sequence) <= 13:
        melting_temperature = (a_count + t_count) * 2 + (g_count + c_count) * 4
    else:
        melting_temperature = 64.9 + 41 * ((g_count + c_count - 16.4) / (a_count + t_count + g_count + c_count))
    return melting_temperature


def salt_adjusted_melting_temperature_calc(sequence, salt):
    """
    Calculate the salt-adjusted melting temperature (Tm) of a primer sequence.

    :param sequence: Primer nucleotides sequence
    :type sequence: str
    :param salt: Sodium ion concentration in moles (unit mM)
    :type salt: float
    :return: Salt-adjusted melting temperature as float
    """
    a_count = sequence.count('A')
    t_count = sequence.count('T')
    c_count = sequence.count('C')
    g_count = sequence.count('G')
    seq_length = len(sequence)
    if seq_length <= 13:
        salt_adjustment = 16.6 * (math.log10(salt) - 3) - 16.6 * math.log10(0.050)
        tm = (a_count + t_count) * 2 + (g_count + c_count) * 4 + salt_adjustment
    else:
        tm = (
            100.5 + (41 * (g_count + c_count) / seq_length)
            - (820 / seq_length)
            + 16.6 * (math.log10(salt) - 3)
        )
    return tm


def gc_clamp_calc(sequence):
    """
    Calculate GC clamp.

    :param sequence: primer sequence
    :type sequence: str
    :return: number of guanine (G) or cytosine (C) bases in the last 5 bases of the primer
    """
    if len(sequence) < 5:
        return 0
    return sequence[-5:].count('G') + sequence[-5:].count('C')


def e260_ssnn_calc(sequence):
    """
    Calculate the extinction coefficient for a primer.

    It uses nearest-neighbor model for single strand DNA (ss-nn) based on https://www.sigmaaldrich.com/US/en/technical-documents/technical-article/genomics/pcr/quantitation-of-oligos.

    :param sequence: primer sequence
    :type sequence: str
    :return: extinction coefficient as float
    """
    e260 = 0
    for i in range(len(sequence) - 1):
        e260 += NN53_EXTINCTION_COEFFICIENTS[sequence[i]][sequence[i + 1]]
    for base in sequence[1:-1]:
        e260 -= BASE_EXTINCTION_COEFFICIENTS[base]
    return e260



from opr import Primer, MeltingTemperature

TEST_CASE_NAME = "Calculations tests"


def test_mwc():
    oprimer = Primer("ATCGATCGATCGATCGAT")
    assert round(oprimer.molecular_weight, 1) == 5498.7


def test_gc_content_1():  # Reference: https://jamiemcgowan.ie/bioinf/gc_content.html
    oprimer = Primer("ATCG")
    assert oprimer.gc_content == 0.5


def test_gc_content_2():  # Reference: https://jamiemcgowan.ie/bioinf/gc_content.html
    oprimer = Primer("ATTCG")
    assert oprimer.gc_content == 0.4


def test_gc_content_3():  # Reference: https://jamiemcgowan.ie/bioinf/gc_content.html
    oprimer = Primer("ATTTTTT")
    assert oprimer.gc_content == 0


def test_gc_clamp_1():  # Reference: https://www.bioinformatics.org/sms2/pcr_primer_stats.html
    oprimer = Primer("ATCGATCGATCGATCGGTCG")
    assert oprimer.gc_clamp == 4


def test_gc_clamp_2():  # Reference: https://www.bioinformatics.org/sms2/pcr_primer_stats.html
    oprimer = Primer("ATCG")
    assert oprimer.gc_clamp == 0


def test_gc_clamp_3():  # Reference: https://www.bioinformatics.org/sms2/pcr_primer_stats.html
    oprimer = Primer("ACTTA")
    assert oprimer.gc_clamp == 1


def test_melt_temp_1():  # Reference: http://biotools.nubic.northwestern.edu/OligoCalc.html
    oprimer = Primer("ATCGATCGATCGATCGATCG")
    basic_melt_temp = oprimer.melting_temperature(MeltingTemperature.BASIC)
    assert round(basic_melt_temp, 1) == 51.8


def test_melt_temp_2():  # Reference: http://biotools.nubic.northwestern.edu/OligoCalc.html
    oprimer = Primer("ATCG")
    basic_melt_temp = oprimer.melting_temperature(method=MeltingTemperature.BASIC)
    assert round(basic_melt_temp, 1) == 12


def test_melt_temp_3():  # Reference: http://biotools.nubic.northwestern.edu/OligoCalc.html
    oprimer = Primer("CTGGAGGACGGAAGAGGAAGTAA")
    salt_adjusted_melt_temp = oprimer.melting_temperature(method=MeltingTemperature.SALT_ADJUSTED)
    assert round(salt_adjusted_melt_temp, 0) == 65.0


def test_melt_temp_4():  # Reference: http://biotools.nubic.northwestern.edu/OligoCalc.html
    oprimer = Primer("CTGGAGGACGGAAGAGGAAGTAAA", salt=65)
    salt_adjusted_melt_temp = oprimer.melting_temperature(method=MeltingTemperature.SALT_ADJUSTED)
    assert round(salt_adjusted_melt_temp, 0) == 67.0


def test_melt_temp_5():  # Reference: http://biotools.nubic.northwestern.edu/OligoCalc.html
    oprimer = Primer("CTGGAGG")
    salt_adjusted_melt_temp = oprimer.melting_temperature(method=MeltingTemperature.SALT_ADJUSTED)
    assert round(salt_adjusted_melt_temp, 0) == 24.0


def test_melt_temp_6():  # Reference: http://biotools.nubic.northwestern.edu/OligoCalc.html
    oprimer = Primer("CTGGAGG", salt=65)
    salt_adjusted_melt_temp = oprimer.melting_temperature(method=MeltingTemperature.SALT_ADJUSTED)
    assert round(salt_adjusted_melt_temp, 0) == 26.0


def test_single_runs_1():  # Reference: https://www.oligoevaluator.com/OligoCalcServlet
    oprimer = Primer("ATCGATCG")
    runs = oprimer.single_runs
    assert runs['A'] == 1 and runs['T'] == 1 and runs['C'] == 1 and runs['G'] == 1


def test_single_runs_2():  # Reference: https://www.oligoevaluator.com/OligoCalcServlet
    oprimer = Primer("ATTCGATCCCCG")
    runs = oprimer.single_runs
    assert runs['A'] == 1 and runs['T'] == 2 and runs['C'] == 4 and runs['G'] == 1


def test_single_runs_3():  # Reference: https://www.oligoevaluator.com/OligoCalcServlet
    oprimer = Primer("AAAAATTCGGGGATCCCCG")
    runs = oprimer.single_runs
    assert runs['A'] == 5 and runs['T'] == 2 and runs['C'] == 4 and runs['G'] == 4


def test_double_runs():
    p1 = Primer("ATATCGAACACACACACA")
    double_runs = p1.double_runs
    true_answer = {
        'AT': 2,
        'AC': 5,
        'AG': 0,
        'TA': 1,
        'TC': 1,
        'TG': 0,
        'CA': 5,
        'CT': 0,
        'CG': 1,
        'GA': 1,
        'GT': 0,
        'GC': 0
    }
    assert len(true_answer) == len(double_runs) and all(double_runs[pair] == true_answer[pair] for pair in double_runs)


def test_repeats_1():
    p = Primer("ATCG")
    assert (
        p.repeats(sequence="A", consecutive=False) == 1 and
        p.repeats(sequence="AT", consecutive=False) == 1 and
        p.repeats(sequence="AC", consecutive=False) == 0 and
        p.repeats(sequence="A", consecutive=True) == 1 and
        p.repeats(sequence="AT", consecutive=True) == 1
    )


def test_repeats_2():
    p = Primer("AAAATCGTGT")
    assert (
        p.repeats(sequence="AA", consecutive=True) == 2 and
        p.repeats(sequence="GT", consecutive=True) == 2
    )


def test_repeats_3():
    p = Primer("ATCGATCGATCG")
    assert p.repeats(sequence="ATCG", consecutive=True) == 3


def test_e260_1():  # https://atdbio.com/tools/oligo-calculator
    oprimer = Primer("ATCGATCGATCGATCGAT")
    assert round(oprimer.E260, 1) == 179.6


def test_e260_2():  # https://atdbio.com/tools/oligo-calculator
    oprimer = Primer("ACGT")
    assert round(oprimer.E260, 1) == 40.3



# -*- coding: utf-8 -*-
"""OPR parameters and constants."""
OPR_VERSION = "0.4"
VALID_BASES = set('ATCG')
DNA_COMPLEMENT_MAP = {"A": "T", "C": "G", "G": "C", "T": "A"}

PRIMER_LOWER_LENGTH = 18
PRIMER_HIGHEST_LENGTH = 30
PRIMER_LOWEST_GC_RANGE = 0.3
PRIMER_HIGHEST_GC_RANGE = 0.8

A_WEIGHT = 313.21
T_WEIGHT = 304.2
C_WEIGHT = 289.18
G_WEIGHT = 329.21
ANHYDROUS_MOLECULAR_WEIGHT_CONSTANT = 61.96

BASE_EXTINCTION_COEFFICIENTS = {  # L ⋅ mmol-1 ⋅ cm-1
    "A": 15.4,
    "C": 7.4,
    "G": 11.5,
    "T": 8.7
}
NN53_EXTINCTION_COEFFICIENTS = {  # L ⋅ mmol-1 ⋅ cm-1
    "A": {
        "A": 27.4,
        "C": 21.2,
        "G": 25.0,
        "T": 22.8
    },
    "C": {
        "A": 21.2,
        "C": 14.6,
        "G": 18.0,
        "T": 15.2
    },
    "G": {
        "A": 25.2,
        "C": 17.6,
        "G": 21.6,
        "T": 20.0
    },
    "T": {
        "A": 23.4,
        "C": 16.2,
        "G": 19.0,
        "T": 16.8
    }
}

DEFAULT_PRIMER_NAME = "unknown"

PRIMER_SEQUENCE_TYPE_ERROR = "Primer sequence should be a string variable."
PRIMER_SEQUENCE_LENGTH_WARNING = "The recommended range for primer length is between 18 and 30."
PRIMER_SEQUENCE_VALID_BASES_ERROR = "Primer sequence should only contain the nucleotide bases A, T, C, and G."
PRIMER_SEQUENCE_VALID_GC_CONTENT_RANGE_WARNING = "The recommended range for GC content is between 30% and 80%."

PRIMER_ADDITION_ERROR = "You can only add two Primer objects."
PRIMER_MULTIPLICATION_ERROR = "The primer sequence can only be multiplied by an integer."

PRIMER_MELTING_TEMPERATURE_NOT_IMPLEMENTED_ERROR = "This method for calculating melting temperature has not been implemented."

PRIMER_ATTRIBUTE_NOT_COMPUTABLE_ERROR = "This attribute either doesn't exist or cannot be computed/cached."



from opr import Primer

TEST_CASE_NAME = "Property tests"


def test_sequence():
    oprimer = Primer("ATCGATCGATCGATCGAT")
    assert oprimer.sequence == "ATCGATCGATCGATCGAT"


def test_name():
    oprimer = Primer("ATCGATCGATCGATCGAT", "primer1")
    assert oprimer.name == "primer1"



# -*- coding: utf-8 -*-
"""Setup module."""
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_requires():
    """Read requirements.txt."""
    requirements = open("requirements.txt", "r").read()
    return list(filter(lambda x: x != "", requirements.split()))


def read_description():
    """Read README.md and CHANGELOG.md."""
    try:
        with open("README.md") as r:
            description = "\n"
            description += r.read()
        with open("CHANGELOG.md") as c:
            description += "\n"
            description += c.read()
        return description
    except Exception:
        return '''OPR is an open-source Python package designed to simplify and streamline primer design and analysis for biologists and bioinformaticians. OPR enables users to design, validate, and optimize primers with ease, catering to a wide range of applications such as PCR, qPCR, and sequencing. With a focus on user-friendliness and efficiency, OPR aims to bridge the gap between biological research and computational tools, making primer-related workflows faster and more reliable.'''


setup(
    name='opr',
    packages=[
        'opr', ],
    version='0.4',
    description='OPR: Optimized Primer',
    long_description=read_description(),
    long_description_content_type='text/markdown',
    author='OPR Development Team',
    author_email='opr@openscilab.com',
    url='https://github.com/openscilab/opr',
    download_url='https://github.com/openscilab/opr/tarball/v0.4',
    keywords="primer biology bioinformatics genome dna pcr",
    project_urls={
            'Source': 'https://github.com/openscilab/opr',
    },
    install_requires=get_requires(),
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Manufacturing',
        'Intended Audience :: Science/Research',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    license='MIT',
)
